@@ -9,7 +9,7 @@
 
 use super::{
     initialize_instance, initialize_vmcontext, InstanceAllocationRequest, InstanceAllocator,
-    InstanceHandle, InstantiationError,
+    InstanceHandle, InstantiationError, ResourceLimiter,
 };
 use crate::{instance::Instance, Memory, Mmap, Table, VMContext};
 use anyhow::{anyhow, bail, Context, Result};
@@ -18,10 +18,11 @@ use std::cell::RefCell;
 use std::cmp::min;
 use std::convert::TryFrom;
 use std::mem;
+use std::rc::Rc;
 use std::sync::{Arc, Mutex};
 use wasmtime_environ::{
     entity::{EntitySet, PrimaryMap},
-    MemoryStyle, Module, Tunables, VMOffsets, WASM_PAGE_SIZE,
+    MemoryStyle, Module, Tunables, VMOffsets, VMOffsetsFields, WASM_PAGE_SIZE,
 };
 
 cfg_if::cfg_if! {
@@ -302,7 +303,7 @@ impl InstancePool {
         let page_size = region::page::size();
 
         // Calculate the maximum size of an Instance structure given the limits
-        let offsets = VMOffsets {
+        let offsets = VMOffsets::from(VMOffsetsFields {
             pointer_size: std::mem::size_of::<*const u8>() as u8,
             num_signature_ids: module_limits.types,
             num_imported_functions: module_limits.imported_functions,
@@ -313,7 +314,7 @@ impl InstancePool {
             num_defined_tables: module_limits.tables,
             num_defined_memories: module_limits.memories,
             num_defined_globals: module_limits.globals,
-        };
+        });
 
         let instance_size = round_up_to_pow2(
             mem::size_of::<Instance>()
@@ -376,10 +377,45 @@ impl InstancePool {
         }
     }
 
+    unsafe fn setup_instance(
+        &self,
+        index: usize,
+        mut req: InstanceAllocationRequest,
+    ) -> Result<InstanceHandle, InstantiationError> {
+        let instance = self.instance(index);
+
+        instance.module = req.module.clone();
+        instance.offsets = VMOffsets::new(
+            std::mem::size_of::<*const u8>() as u8,
+            instance.module.as_ref(),
+        );
+        instance.host_state = std::mem::replace(&mut req.host_state, Box::new(()));
+
+        Self::set_instance_memories(
+            instance,
+            self.memories.get(index),
+            self.memories.max_wasm_pages,
+            req.limiter,
+        )?;
+
+        Self::set_instance_tables(
+            instance,
+            self.tables.get(index),
+            self.tables.max_elements,
+            req.limiter,
+        )?;
+
+        initialize_vmcontext(instance, req);
+
+        Ok(InstanceHandle {
+            instance: instance as _,
+        })
+    }
+
     fn allocate(
         &self,
         strategy: PoolingAllocationStrategy,
-        mut req: InstanceAllocationRequest,
+        req: InstanceAllocationRequest,
     ) -> Result<InstanceHandle, InstantiationError> {
         let index = {
             let mut free_list = self.free_list.lock().unwrap();
@@ -390,28 +426,15 @@ impl InstancePool {
             free_list.swap_remove(free_index)
         };
 
-        let host_state = std::mem::replace(&mut req.host_state, Box::new(()));
-
         unsafe {
-            let instance = self.instance(index);
-
-            instance.module = req.module.clone();
-            instance.offsets = VMOffsets::new(
-                std::mem::size_of::<*const u8>() as u8,
-                instance.module.as_ref(),
-            );
-            instance.host_state = host_state;
-
-            Self::set_instance_memories(
-                instance,
-                self.memories.get(index),
-                self.memories.max_wasm_pages,
-            )?;
-            Self::set_instance_tables(instance, self.tables.get(index), self.tables.max_elements)?;
-
-            initialize_vmcontext(instance, req);
-
-            Ok(InstanceHandle::new(instance as _))
+            self.setup_instance(index, req).or_else(|e| {
+                // Deallocate the allocated instance on error
+                let instance = self.instance(index);
+                self.deallocate(&InstanceHandle {
+                    instance: instance as _,
+                });
+                Err(e)
+            })
         }
     }
 
@@ -473,6 +496,7 @@ impl InstancePool {
         instance: &mut Instance,
         mut memories: impl Iterator<Item = *mut u8>,
         max_pages: u32,
+        limiter: Option<&Rc<dyn ResourceLimiter>>,
     ) -> Result<(), InstantiationError> {
         let module = instance.module.as_ref();
 
@@ -487,6 +511,7 @@ impl InstancePool {
                     memories.next().unwrap(),
                     max_pages,
                     commit_memory_pages,
+                    limiter,
                 )
                 .map_err(InstantiationError::Resource)?,
             );
@@ -503,6 +528,7 @@ impl InstancePool {
         instance: &mut Instance,
         mut tables: impl Iterator<Item = *mut u8>,
         max_elements: u32,
+        limiter: Option<&Rc<dyn ResourceLimiter>>,
     ) -> Result<(), InstantiationError> {
         let module = instance.module.as_ref();
 
@@ -514,9 +540,10 @@ impl InstancePool {
             commit_table_pages(base, max_elements as usize * mem::size_of::<*mut u8>())
                 .map_err(InstantiationError::Resource)?;
 
-            instance
-                .tables
-                .push(Table::new_static(plan, base as _, max_elements));
+            instance.tables.push(
+                Table::new_static(plan, base as _, max_elements, limiter)
+                    .map_err(InstantiationError::Resource)?,
+            );
         }
 
         let mut dropped_elements = instance.dropped_elements.borrow_mut();
@@ -1341,7 +1368,9 @@ mod test {
         assert_eq!(instances.offsets.num_defined_tables, 1);
         assert_eq!(instances.offsets.num_defined_memories, 1);
         assert_eq!(instances.offsets.num_defined_globals, 0);
-        assert_eq!(instances.instance_size, 4096);
+        // As of April 2021, the instance struct's size is largely below the size of a single page,
+        // so it's safe to assume it's been rounded to the size of a single memory page here.
+        assert_eq!(instances.instance_size, region::page::size());
         assert_eq!(instances.max_instances, 3);
 
         assert_eq!(&*instances.free_list.lock().unwrap(), &[0, 1, 2]);
@@ -1364,11 +1393,12 @@ mod test {
                                 memories: &[],
                                 globals: &[],
                             },
-                            lookup_shared_signature: &|_| VMSharedSignatureIndex::default(),
+                            shared_signatures: VMSharedSignatureIndex::default().into(),
                             host_state: Box::new(()),
                             interrupts: std::ptr::null(),
                             externref_activations_table: std::ptr::null_mut(),
-                            stack_map_registry: std::ptr::null_mut(),
+                            module_info_lookup: None,
+                            limiter: None,
                         },
                     )
                     .expect("allocation should succeed"),
@@ -1388,11 +1418,12 @@ mod test {
                     memories: &[],
                     globals: &[],
                 },
-                lookup_shared_signature: &|_| VMSharedSignatureIndex::default(),
+                shared_signatures: VMSharedSignatureIndex::default().into(),
                 host_state: Box::new(()),
                 interrupts: std::ptr::null(),
                 externref_activations_table: std::ptr::null_mut(),
-                stack_map_registry: std::ptr::null_mut(),
+                module_info_lookup: None,
+                limiter: None,
             },
         ) {
             Err(InstantiationError::Limit(3)) => {}
@@ -1477,10 +1508,12 @@ mod test {
             },
         )?;
 
-        assert_eq!(pool.table_size, 4096);
+        let host_page_size = region::page::size();
+
+        assert_eq!(pool.table_size, host_page_size);
         assert_eq!(pool.max_tables, 4);
         assert_eq!(pool.max_instances, 7);
-        assert_eq!(pool.page_size, 4096);
+        assert_eq!(pool.page_size, host_page_size);
         assert_eq!(pool.max_elements, 100);
 
         let base = pool.mapping.as_ptr() as usize;
@@ -1512,9 +1545,10 @@ mod test {
             1,
         )?;
 
-        assert_eq!(pool.stack_size, 8192);
+        let native_page_size = region::page::size();
+        assert_eq!(pool.stack_size, 2 * native_page_size);
         assert_eq!(pool.max_instances, 10);
-        assert_eq!(pool.page_size, 4096);
+        assert_eq!(pool.page_size, native_page_size);
 
         assert_eq!(
             &*pool.free_list.lock().unwrap(),
@@ -1576,7 +1610,7 @@ mod test {
     }
 
     #[test]
-    fn test_pooling_allocator_with_memory_pages_exeeded() {
+    fn test_pooling_allocator_with_memory_pages_exceeded() {
         assert_eq!(
             PoolingInstanceAllocator::new(
                 PoolingAllocationStrategy::Random,
@@ -1597,7 +1631,7 @@ mod test {
     }
 
     #[test]
-    fn test_pooling_allocator_with_reservation_size_exeeded() {
+    fn test_pooling_allocator_with_reservation_size_exceeded() {
         assert_eq!(
             PoolingInstanceAllocator::new(
                 PoolingAllocationStrategy::Random,
