@@ -519,11 +519,12 @@ mod test {
                                     memories: &[],
                                     globals: &[],
                                 },
-                                lookup_shared_signature: &|_| VMSharedSignatureIndex::default(),
+                                shared_signatures: VMSharedSignatureIndex::default().into(),
                                 host_state: Box::new(()),
                                 interrupts: ptr::null(),
                                 externref_activations_table: ptr::null_mut(),
-                                stack_map_registry: ptr::null_mut(),
+                                module_info_lookup: None,
+                                limiter: None,
                             },
                         )
                         .expect("instance should allocate"),
